/***************************************************************************
 *   Copyright (C) 2009 by Saïd LANKRI   *
 *   said.lankri@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "QuantumStyle.h"

#include <stdlib.h>
#include <limits.h>

#include <QApplication>
#include <QWidget>
#include <QPainter>
#include <QRect>
#include <QSvgRenderer>
#include <QStyleOption>
#include <QSettings>
#include <QFile>
#include <QVariant>
#include <QBoxLayout>
#include <QGridLayout>
#include <QFont>

#include <QSpinBox>
#include <QToolButton>
#include <QToolBar>
#include <QPushButton>
#include <QComboBox>

#include "../themeconfig/ThemeConfig.h"

//#define __DEBUG__

#ifdef __DEBUG__
#define DEBUG(...) printf("QuantumStyle "__VA_ARGS__)
#else
#define DEBUG(...)
#endif

#define MAX(X,Y) (X) > (Y) ? (X) : (Y)

QuantumStyle::QuantumStyle()
  : QCommonStyle()
{
  DEBUG("> INIT\n");
  bool fr = false;

  defaultRndr = new QSvgRenderer(this);
  themeRndr = new QSvgRenderer(this);
  appRndr = new QSvgRenderer(this);

  defaultSettings = themeSettings = appSettings = NULL;
  globalSettings = NULL;

  char *xdg_config_home = getenv("XDG_CONFIG_HOME");
  if (!xdg_config_home) {
	char *home = getenv("HOME");
	if (!home)
	  qDebug("CRITICAL : HOME environment variable not found ! You'll have problems !");

	fr = true;
    xdg_config_home = strdup(QString("%1/.config").arg(home).toAscii());
  }

  // load global config file
  if (QFile::exists(QString("%1/QuantumStyle/quantumstyle.qsconfig").arg(xdg_config_home)))
    globalSettings = new QSettings(QString("%1/QuantumStyle/quantumstyle.qsconfig").arg(xdg_config_home),QSettings::NativeFormat);

  if (globalSettings && globalSettings->contains("theme"))
    m_theme = globalSettings->value("theme").toString();

  // load theme settings
  // default theme
  defaultSettings = new ThemeConfig(":default.qsconfig");

  // user chosen theme
  if (!m_theme.isEmpty() && QFile::exists(QString("%1/QuantumStyle/%2/%2.qsconfig").arg(xdg_config_home).arg(m_theme)))
    themeSettings = new ThemeConfig(QString("%1/QuantumStyle/%2/%2.qsconfig").arg(xdg_config_home).arg(m_theme));

  // application specific theme
  if (QFile::exists(QString("%1/.QuantumStyle/%2/%2.qsconfig").arg(xdg_config_home).arg(QApplication::applicationName())))
    appSettings = new ThemeConfig(QString("%1/QuantumStyle/%2/%2.qsconfig").arg(xdg_config_home).arg(QApplication::applicationName()));

  // load SVG renderers for default, user and application theme
  defaultRndr->load(QString(":default.svg"));

  if (QFile::exists(QString("%1/QuantumStyle/%2/%2.svg").arg(xdg_config_home).arg(m_theme)))
	themeRndr->load(QString("%1/QuantumStyle/%2/%2.svg").arg(xdg_config_home).arg(m_theme));

  if (QFile::exists(QString("%1/QuantumStyle/%2/%2.svg").arg(xdg_config_home).arg(QApplication::applicationName())))
	appRndr->load(QString("%1/QuantumStyle/%2/%2.svg").arg(xdg_config_home).arg(QApplication::applicationName()));

  if (fr)
	free(xdg_config_home);

  DEBUG("< INIT\n");
}

QuantumStyle::~QuantumStyle()
{
  delete defaultSettings;
  delete themeSettings;
  delete appSettings;

  delete defaultRndr;
  delete appRndr;
  delete themeRndr;
}

void QuantumStyle::drawPrimitive(PrimitiveElement element, const QStyleOption * option, QPainter * painter, const QWidget * widget) const
{
  DEBUG("> drawPrimitive\n");

  int x,y,h,w;

  option->rect.getRect(&x,&y,&w,&h);
  const QString status =
        (option->state & State_Enabled) ?
          (option->state & State_Sunken) ? "pressed" :
          (option->state & State_On) ? "toggled" :
          (option->state & State_Selected) ? "toggled" :
          (option->state & State_MouseOver) ? "focused" : "normal"
        : "disabled";

  #ifdef __DEBUG__
  if (widget) {
    DEBUG("********** widget %s\n",(const char *)(widget->objectName().toAscii()));
  }
  #endif

  switch(element) {
    case PE_Widget : {
      DEBUG("> PE_Widget\n");

      QString group = "PE_Widget";
      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      DEBUG("< PE_Widget\n");
      break;
    }

    case PE_PanelButtonCommand : {
      DEBUG("> PE_PanelButtonCommand\n");

      QString group = "PE_PanelButtonCommand";

      frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      if (fspec.hasCapsule)
        capsulePosition(widget,fspec.hasCapsule,fspec.capsuleH,fspec.capsuleV);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      DEBUG("< PE_PanelButtonCommand\n");
      break;
    }

    case PE_PanelButtonTool : {
      DEBUG("PE_PanelButtonTool\n");
      QString group = "PE_PanelButtonTool";

      frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      const QToolButton *w = qobject_cast<const QToolButton *>(widget);

      if (fspec.hasCapsule)
        capsulePosition(widget,fspec.hasCapsule,fspec.capsuleH,fspec.capsuleV);

      if (w) {
        if ( w->autoRaise() && ( (status == "normal") || (status == "disabled") ) ) {
          ;
        } else {
          renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
        }
      } else {
        renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      }

      DEBUG("PE_PanelButtonTool\n");
      break;
    }

    case PE_IndicatorRadioButton : {
      const QString group = "PE_RadioButton";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      if (option->state & State_Enabled)
        if (option->state & State_MouseOver)
          if (option->state & State_On) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-checked-focused");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-focused");
          } else {
            renderFrame(painter,option->rect,fspec,fspec.element+"-focused");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-focused");
          }
        else
          if (option->state & State_On) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-checked-normal");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-normal");
          } else {
            renderFrame(painter,option->rect,fspec,fspec.element+"-normal");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-normal");
          }
      else
        if (option->state & State_On) {
          renderFrame(painter,option->rect,fspec,fspec.element+"-checked-disabled");
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-disabled");
        } else {
          renderFrame(painter,option->rect,fspec,fspec.element+"-disabled");
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-disabled");
        }

      break;
    }

    case PE_IndicatorCheckBox : {
      const QString group = "PE_CheckBox";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      if (option->state & State_Enabled)
        if (option->state & State_MouseOver)
          if (option->state & State_On) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-checked-focused");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-focused");
          } else if (option->state & State_NoChange) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-tristate-focused");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-tristate-focused");
          } else {
            renderFrame(painter,option->rect,fspec,fspec.element+"-focused");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-focused");
          }
        else
          if (option->state & State_On) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-checked-normal");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-normal");
          } else if (option->state & State_NoChange) {
            renderFrame(painter,option->rect,fspec,fspec.element+"-tristate-normal");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-tristate-normal");
          } else {
            renderFrame(painter,option->rect,fspec,fspec.element+"-normal");
            renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-normal");
          }
      else
        if (option->state & State_On) {
          renderFrame(painter,option->rect,fspec,fspec.element+"-checked-disabled");
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-checked-disabled");
        } else  if (option->state & State_NoChange) {
          renderFrame(painter,option->rect,fspec,fspec.element+"-tristate-disabled");
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-tristate-disabled");
        } else {
          renderFrame(painter,option->rect,fspec,fspec.element+"-disabled");
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-disabled");
        }

      break;
    }

    case PE_FrameFocusRect : {
      const QString group = "PE_Focus";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element);

      break;
    }

    case PE_IndicatorBranch : {
      QString group = "PE_TreeExpander";

      frame_spec_t fspec = getFrameSpec(group);
      interior_spec_t ispec = getInteriorSpec(group);
      indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      if (option->state & State_Children) {
        if (option->state & State_Open)
          renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-minus-"+status);
        else
          renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-plus-"+status);
      }

      break;
    }

    case PE_FrameMenu :  {
      const QString group = "PE_MenuItem";

      const frame_spec_t fspec = getFrameSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-toggled");

      break;
    }

    case PE_FrameWindow : {
      const QString group = "PE_GenericFrame";

      const frame_spec_t fspec = getFrameSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);

      break;
    }

    case PE_Frame :
    case PE_FrameGroupBox :
    case PE_FrameDockWidget :
    case PE_FrameButtonBevel :
    case PE_FrameStatusBar :
    case PE_FrameTabWidget :
    case PE_FrameTabBarBase : {
      const QString group = "PE_GenericFrame";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case PE_FrameLineEdit : {
      const QString group = "PE_LineEdit";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);

      break;
    }

    case PE_PanelLineEdit : {
      const QString group = "PE_LineEdit";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case PE_PanelToolBar : {
      const QString group = "PE_Toolbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case PE_IndicatorToolBarHandle : {
      const QString group = "PE_Toolbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderInterior(painter,option->rect,fspec,ispec,dspec.element+"-handle-"+status);

      break;
    }

    case PE_IndicatorToolBarSeparator : {
      const QString group = "PE_Toolbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderInterior(painter,option->rect,fspec,ispec,dspec.element+"-separator-"+status);

      break;
    }

    case PE_IndicatorSpinPlus : {
      const QString group = "PE_IndicatorSpinBox";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-plus-"+status);

      break;
    }

    case PE_IndicatorSpinMinus : {
      const QString group = "PE_IndicatorSpinBox";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-minus-"+status);

      break;
    }

    case PE_IndicatorSpinUp : {
      const QString group = "PE_IndicatorSpinBox";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-up-"+status);

      break;
    }

    case PE_IndicatorSpinDown : {
      const QString group = "PE_IndicatorSpinBox";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-down-"+status);

      break;
    }

    case PE_IndicatorHeaderArrow : {
      const QString group = "PE_IndicatorHeaderArrow";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      if (option->state & State_DownArrow)
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-down-"+status);
      else
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-up-"+status);

      break;
    }

    case PE_IndicatorButtonDropDown : {
      const QString group = "PE_DropDownButton";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      const QToolButton *w = qobject_cast<const QToolButton *>(widget);

      if (w) {
        if ( w->autoRaise() && ( (status == "normal") || (status == "disabled") ) ) {
          ;
        } else {
          renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
        }
      } else {
        renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      }

      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-"+status);

      break;
    }

    case PE_PanelMenuBar : {
      break;
    }

    case PE_IndicatorTabTear : {
      const QString group = "PE_Tab";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderInterior(painter,option->rect,fspec,ispec,dspec.element+"-tear-"+status);

      break;
    }

    case PE_IndicatorArrowUp : {
      const QString group = "PE_IndicatorArrow";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-up-"+status);

      break;
    }

    case PE_IndicatorArrowDown : {
      const QString group = "PE_IndicatorArrow";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-down-"+status);

      break;
    }

    case PE_IndicatorArrowLeft : {
      const QString group = "PE_IndicatorArrow";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-left-"+status);

      break;
    }

    case PE_IndicatorArrowRight : {
      const QString group = "PE_IndicatorArrow";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-right-"+status);

      break;
    }

    case PE_IndicatorProgressChunk : {
      const QString group = "PE_ProgressbarContents";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case PE_IndicatorDockWidgetResizeHandle : {
      drawControl(CE_Splitter,option,painter,widget);

      break;
    }

    case PE_IndicatorMenuCheckMark : {
      // nothing, uses radio and checkbox

      break;
    }

    default :
      QCommonStyle::drawPrimitive(element,option,painter,widget);
      break;
  }

  DEBUG("< drawPrimitive\n");
}

void QuantumStyle::drawControl(ControlElement element, const QStyleOption * option, QPainter * painter, const QWidget * widget) const
{
  DEBUG("> drawControl\n");

  int x,y,h,w;
  option->rect.getRect(&x,&y,&w,&h);
  const QString status =
      (option->state & State_Enabled) ?
        (option->state & State_Sunken) ? "pressed" :
        (option->state & State_On) ? "toggled" :
        (option->state & State_Selected) ? "toggled" :
        (option->state & State_MouseOver) ? "focused" : "normal"
      : "disabled";

  const QIcon::Mode iconmode =
        (option->state & State_Enabled) ?
        (option->state & State_Sunken) ? QIcon::Active :
        (option->state & State_MouseOver) ? QIcon::Active : QIcon::Normal
      : QIcon::Disabled;

  const QIcon::State iconstate =
      (option->state & State_On) ? QIcon::On : QIcon::Off;

  #ifdef __DEBUG__
  if (widget) {
    DEBUG("********** widget %s\n",(const char *)(widget->objectName().toAscii()));
  }
  #endif

  switch (element) {
    case CE_MenuItem : {
      const QStyleOptionMenuItem *opt =
          qstyleoption_cast<const QStyleOptionMenuItem *>(option);

      if (opt) {
        const QString group = "PE_MenuItem";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const indicator_spec_t dspec = getIndicatorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);

        if (opt->menuItemType == QStyleOptionMenuItem::Separator)
          renderElement(painter,dspec.element+"-separator",option->rect,true,20,0);
        else if (opt->menuItemType == QStyleOptionMenuItem::TearOff)
          renderElement(painter,dspec.element+"-tearoff",option->rect,true,20,0);
        else {
          renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
          renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

          const QStringList l = opt->text.split('\t');

          if (l.size() > 0) {
            // menu label
            if (opt->icon.isNull())
              renderLabel(painter,option->rect.adjusted(opt->maxIconWidth+lspec.tispace,0,0,0),fspec,ispec,lspec,Qt::AlignLeft|Qt::AlignVCenter | Qt::TextShowMnemonic |Qt::TextSingleLine,l[0]);
            else
              renderLabel(painter,option->rect,fspec,ispec,lspec,Qt::AlignLeft|Qt::AlignVCenter | Qt::TextShowMnemonic |Qt::TextSingleLine,l[0],opt->icon.pixmap(opt->maxIconWidth,iconmode,iconstate));
          }
          if (l.size() > 1)
            // shortcut
            renderLabel(painter,option->rect.adjusted(opt->maxIconWidth,0,-15,0),fspec,ispec,lspec,Qt::AlignRight|Qt::AlignVCenter | Qt::TextShowMnemonic| Qt::TextSingleLine,l[1]);

          QStyleOptionMenuItem o(*opt);
          o.rect = alignedRect(QApplication::layoutDirection(),Qt::AlignRight | Qt::AlignVCenter,QSize(10,10),textRect(o.rect,fspec,ispec,lspec));
          if (opt->menuItemType == QStyleOptionMenuItem::SubMenu) {
            drawPrimitive(PE_IndicatorArrowRight,&o,painter);
          }

          if (opt->checkType == QStyleOptionMenuItem::Exclusive) {
            if (opt->checked)
              o.state |= State_On;
            drawPrimitive(PE_IndicatorRadioButton,&o,painter,widget);
          }

          if (opt->checkType == QStyleOptionMenuItem::NonExclusive) {
            if (opt->checked)
              o.state |= State_On;
            drawPrimitive(PE_IndicatorCheckBox,&o,painter,widget);
          }
        }
      }

      break;
    }

    case CE_MenuEmptyArea : {
      break;
    }

    case CE_MenuBarItem : {
      const QStyleOptionMenuItem *opt =
          qstyleoption_cast<const QStyleOptionMenuItem *>(option);

      if (opt) {
        QString group = "PE_MenuBar";
        frame_spec_t fspec = getFrameSpec(group);
        fspec.left = fspec.right = 0;
        group = "PE_MenuBarItem";
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);

        renderFrame(painter,option->rect,fspec,fspec.element);
        renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

        renderLabel(painter,option->rect,fspec,ispec,lspec,Qt::AlignLeft | Qt::AlignVCenter | Qt::TextShowMnemonic | Qt::TextSingleLine,opt->text);
      }

      break;
    }

    case CE_MenuBarEmptyArea : {
        const QString group = "PE_MenuBar";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);

        // NOTE this does not use the status (otherwise always disabled)
        renderFrame(painter,option->rect,fspec,fspec.element);
        renderInterior(painter,option->rect,fspec,ispec,ispec.element);

      break;
    }

    case CE_MenuScroller : {
      drawPrimitive(PE_PanelButtonTool,option,painter,widget);
      if (option->state & State_DownArrow)
        drawPrimitive(PE_IndicatorArrowDown,option,painter,widget);
      else
        drawPrimitive(PE_IndicatorArrowUp,option,painter,widget);

      break;
    }

    case CE_MenuHMargin:
    case CE_MenuVMargin:
      break;

    case CE_RadioButtonLabel : {
      const QStyleOptionButton *opt =
          qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        QStyleOptionButton o(*opt);

        const QString group = "PE_RadioButton";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignLeft | Qt::AlignVCenter | Qt::TextShowMnemonic,o.text,opt->icon.pixmap(opt->iconSize,iconmode,iconstate));
      }

      break;
    }

    case CE_CheckBoxLabel : {
      const QStyleOptionButton *opt =
          qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        QStyleOptionButton o(*opt);

        const QString group = "PE_CheckBox";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignLeft | Qt::AlignVCenter | Qt::TextShowMnemonic,o.text,opt->icon.pixmap(opt->iconSize,iconmode,iconstate));
      }

      break;
    }

    case CE_ComboBoxLabel : { // non editable
      const QStyleOptionComboBox *opt =
          qstyleoption_cast<const QStyleOptionComboBox *>(option);

      if (opt && !opt->editable) {
        const QString group = "PE_ComboBox";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        const QRect r = subControlRect(CC_ComboBox,opt,SC_ComboBoxEditField,widget);
        renderLabel(painter,r,fspec,ispec,tspec,Qt::AlignCenter | Qt::AlignVCenter | Qt::TextShowMnemonic,opt->currentText,opt->currentIcon.pixmap(opt->iconSize,iconmode,iconstate));
      }

      break;
    }

    case CE_TabBarTabShape : {
      const QString group = "PE_Tab";

      frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_TabBarTabLabel : {
      const QStyleOptionTab *opt =
        qstyleoption_cast<const QStyleOptionTab *>(option);

      if (opt) {
        const QString group = "PE_Tab";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,lspec,Qt::AlignLeft | Qt::TextShowMnemonic,opt->text,opt->icon.pixmap(pixelMetric(PM_TabBarIconSize),iconmode,iconstate));
      }

      break;
    }

    case CE_ToolBoxTabShape : {
      const QString group = "PE_ToolboxTab";

      frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_ToolBoxTabLabel : {
      const QStyleOptionToolBox *opt =
          qstyleoption_cast<const QStyleOptionToolBox *>(option);

      if (opt) {
        const QString group = "PE_ToolboxTab";

        frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignCenter | Qt::TextShowMnemonic,opt->text,opt->icon.pixmap(pixelMetric(PM_TabBarIconSize),iconmode,iconstate));
      }
    }

    case CE_ProgressBarGroove : {
      const QString group = "PE_Progressbar";

      frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_ProgressBarContents : {
      const QStyleOptionProgressBar *opt =
          qstyleoption_cast<const QStyleOptionProgressBar *>(option);

      if (opt) {
        const QString group = "PE_ProgressbarContents";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);

        int empty = sliderPositionFromValue(opt->minimum,opt->maximum,opt->maximum-opt->progress,w,false);

        renderFrame(painter,option->rect.adjusted(0,0,-empty,0),fspec,fspec.element+"-"+status);
        renderInterior(painter,option->rect.adjusted(0,0,-empty,0),fspec,ispec,ispec.element+"-"+status);
      }

      break;
    }

    case CE_ProgressBarLabel : {
      const QStyleOptionProgressBar *opt =
          qstyleoption_cast<const QStyleOptionProgressBar *>(option);

      if (opt && opt->textVisible) {
        const QString group = "PE_Progressbar";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,tspec,opt->textAlignment,opt->text);
      }

      break;
    }

    case CE_Splitter : {
      const QString group = "PE_Splitter";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_ScrollBarAddLine : {
      const QString group = "PE_Scrollbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      if (option->state & State_Horizontal)
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-right-"+status);
      else
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-down-"+status);

      break;
    }

    case CE_ScrollBarSubLine : {
      const QString group = "PE_Scrollbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      if (option->state & State_Horizontal)
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-left-"+status);
      else
        renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-up-"+status);

      break;
    }

    case CE_ScrollBarSlider : {
      const QString group = "PE_ScrollbarSlider";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_HeaderSection : {
      const QString group = "PE_HeaderSection";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_HeaderLabel : {
      const QStyleOptionHeader *opt =
        qstyleoption_cast<const QStyleOptionHeader *>(option);

      if (opt) {
        const QString group = "PE_HeaderSection";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);

        renderLabel(painter,option->rect,fspec,ispec,lspec,opt->textAlignment,opt->text,opt->icon.pixmap(pixelMetric(PM_SmallIconSize),iconmode,iconstate));
      }

      break;
    }

    case CE_ToolBar : {
      const QString group = "PE_Toolbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);

      break;
    }

    case CE_SizeGrip : {
      const QString group = "PE_SizeGrip";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const indicator_spec_t dspec = getIndicatorSpec(group);

      renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
      renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
      renderIndicator(painter,option->rect,fspec,ispec,dspec,dspec.element+"-"+status);

      break;
    }

    case CE_PushButtonLabel : {
      const QStyleOptionButton *opt =
          qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        const QString group = "PE_PanelButtonCommand";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        const QPushButton *w = qobject_cast<const QPushButton *>(widget);
        if (w && w->isDefault()) {
          QFont f(w->font());
          f.setBold(true);
          painter->setFont(f);
        }

        renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignCenter | Qt::AlignVCenter | Qt::TextShowMnemonic,opt->text,opt->icon.pixmap(opt->iconSize,iconmode,iconstate));
      }

      break;
    }

    case CE_ToolButtonLabel : {
      const QStyleOptionToolButton *opt =
          qstyleoption_cast<const QStyleOptionToolButton *>(option);

      if (opt) {
        const QString group = "PE_PanelButtonTool";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const indicator_spec_t dspec = getIndicatorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        if (opt->arrowType == Qt::NoArrow)
          renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignCenter | Qt::TextShowMnemonic,opt->text,opt->icon.pixmap(opt->iconSize,iconmode,iconstate),opt->toolButtonStyle);
        else
          renderLabel(painter,option->rect.adjusted(dspec.size+tspec.tispace,0,0,0),fspec,ispec,tspec,Qt::AlignCenter | Qt::TextShowMnemonic,opt->text,opt->icon.pixmap(opt->iconSize,iconmode,iconstate),opt->toolButtonStyle);

        switch (opt->arrowType) {
          case Qt::NoArrow :
            break;
          case Qt::UpArrow :
            renderIndicator(painter,textRect(option->rect,fspec,ispec,tspec),fspec,ispec,dspec,dspec.element+"-up-"+status,Qt::AlignLeft | Qt::AlignVCenter);
            break;
          case Qt::DownArrow :
            renderIndicator(painter,textRect(option->rect,fspec,ispec,tspec),fspec,ispec,dspec,dspec.element+"-down-"+status,Qt::AlignLeft | Qt::AlignVCenter);
            break;
          case Qt::LeftArrow :
            renderIndicator(painter,textRect(option->rect,fspec,ispec,tspec),fspec,ispec,dspec,dspec.element+"-left-"+status,Qt::AlignLeft | Qt::AlignVCenter);
            break;
          case Qt::RightArrow :
            renderIndicator(painter,textRect(option->rect,fspec,ispec,tspec),fspec,ispec,dspec,dspec.element+"-right-"+status,Qt::AlignLeft | Qt::AlignVCenter);
            break;
        }
      }

      break;
    }

    case CE_DockWidgetTitle : {
      const QStyleOptionDockWidget *opt =
          qstyleoption_cast<const QStyleOptionDockWidget *>(option);

      if (opt) {
        const QString group = "PE_Dock";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t tspec = getLabelSpec(group);

        renderFrame(painter,option->rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,option->rect,fspec,ispec,ispec.element+"-"+status);
        renderLabel(painter,option->rect,fspec,ispec,tspec,Qt::AlignLeft | Qt::AlignVCenter | Qt::TextShowMnemonic,opt->title);
      }

      break;
    }

    default : QCommonStyle::drawControl(element,option,painter,widget);
  }

  DEBUG("< drawControl\n");
}

void QuantumStyle::drawComplexControl(ComplexControl control, const QStyleOptionComplex * option, QPainter * painter, const QWidget * widget) const
{
  DEBUG("> drawComplexControl\n");

  int x,y,h,w;
  option->rect.getRect(&x,&y,&w,&h);
  const QString status =
        (option->state & State_Enabled) ?
          (option->state & State_Sunken) ? "pressed" :
          (option->state & State_On) ? "toggled" :
          (option->state & State_Selected) ? "toggled" :
          (option->state & State_MouseOver) ? "focused" : "normal"
        : "disabled";

  #ifdef __DEBUG__
  if (widget) {
    DEBUG("********** widget %s\n",(const char *)(widget->objectName().toAscii()));
  }
  #endif

  switch (control) {
    case CC_ToolButton : {
      const QStyleOptionToolButton *opt =
        qstyleoption_cast<const QStyleOptionToolButton *>(option);

      if (opt) {
        QStyleOptionToolButton o(*opt);

        o.rect = subControlRect(CC_ToolButton,opt,SC_ToolButton,widget);
        drawPrimitive(PE_PanelButtonTool,&o,painter,widget);
        drawControl(CE_ToolButtonLabel,&o,painter,widget);

        if (widget) {
          const QToolButton * w = qobject_cast<const QToolButton *>(widget);
          if (w && w->popupMode() == QToolButton::MenuButtonPopup) {
            o.rect = subControlRect(CC_ToolButton,opt,SC_ToolButtonMenu,widget);
            drawPrimitive(PE_IndicatorButtonDropDown,&o,painter,widget);
          }
        }
      }

      break;
    }

    case CC_SpinBox : {
      const QStyleOptionSpinBox *opt =
        qstyleoption_cast<const QStyleOptionSpinBox *>(option);

      if (opt) {
        QStyleOptionSpinBox o(*opt);

        o.rect = subControlRect(CC_SpinBox,opt,SC_SpinBoxEditField,widget);
        drawPrimitive(PE_FrameLineEdit,&o,painter,widget);

        if (opt->buttonSymbols == QAbstractSpinBox::UpDownArrows) {
          o.rect = subControlRect(CC_SpinBox,opt,SC_SpinBoxUp,widget);
          drawPrimitive(PE_IndicatorSpinUp,&o,painter,widget);
          o.rect = subControlRect(CC_SpinBox,opt,SC_SpinBoxDown,widget);
          drawPrimitive(PE_IndicatorSpinDown,&o,painter,widget);
        }
        if (opt->buttonSymbols == QAbstractSpinBox::PlusMinus) {
          o.rect = subControlRect(CC_SpinBox,opt,SC_SpinBoxUp,widget);
          drawPrimitive(PE_IndicatorSpinPlus,&o,painter,widget);
          o.rect = subControlRect(CC_SpinBox,opt,SC_SpinBoxDown,widget);
          drawPrimitive(PE_IndicatorSpinMinus,&o,painter,widget);
        }
      }

      break;
    }

    case CC_ComboBox : {
      const QStyleOptionComboBox *opt =
          qstyleoption_cast<const QStyleOptionComboBox *>(option);

      if (opt) {
        QStyleOptionComboBox o(*opt);

        QString group = "PE_ComboBox";

        frame_spec_t fspec = getFrameSpec(group);
        interior_spec_t ispec = getInteriorSpec(group);
        label_spec_t lspec = getLabelSpec(group);

        o.rect = subControlRect(CC_ComboBox,opt,SC_ComboBoxEditField,widget);

        renderFrame(painter,o.rect,fspec,fspec.element+"-"+status);

        if (opt->editable)
          drawPrimitive(PE_FrameLineEdit,&o,painter,widget);
        else
          renderInterior(painter,o.rect,fspec,ispec,ispec.element+"-"+status);

        o.rect = subControlRect(CC_ComboBox,opt,SC_ComboBoxArrow,widget);
        drawPrimitive(PE_IndicatorButtonDropDown,&o,painter,widget);
      }

      break;
    }

    case CC_ScrollBar : {
      const QStyleOptionSlider *opt =
        qstyleoption_cast<const QStyleOptionSlider *>(option);

      if (opt) {
        QStyleOptionSlider o(*opt);

        const QString group = "PE_ScrollbarGroove";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);

        o.rect = subControlRect(CC_ScrollBar,opt,SC_ScrollBarGroove,widget);
        renderFrame(painter,o.rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,o.rect,fspec,ispec,ispec.element+"-"+status);

        o.rect = subControlRect(CC_ScrollBar,opt,SC_ScrollBarAddLine,widget);
        drawControl(CE_ScrollBarAddLine,&o,painter,widget);

        o.rect = subControlRect(CC_ScrollBar,opt,SC_ScrollBarSubLine,widget);
        drawControl(CE_ScrollBarSubLine,&o,painter,widget);

        o.rect = subControlRect(CC_ScrollBar,opt,SC_ScrollBarSlider,widget);
        drawControl(CE_ScrollBarSlider,&o,painter,widget);
      }

      break;
    }

    case CC_Slider : {
      const QStyleOptionSlider *opt =
        qstyleoption_cast<const QStyleOptionSlider *>(option);

      if (opt) {
        QStyleOptionSlider o(*opt);

        QString group = "PE_Slider";

        frame_spec_t fspec = getFrameSpec(group);
        interior_spec_t ispec = getInteriorSpec(group);

        QRect empty = subControlRect(CC_Slider,opt,SC_SliderGroove,widget);
        QRect full = subControlRect(CC_Slider,opt,SC_SliderGroove,widget);
        QRect slider = subControlRect(CC_Slider,opt,SC_SliderHandle,widget);

        // take into account inversion
        if (option->state & State_Horizontal)
          if (!opt->upsideDown) {
            full.setWidth(slider.x());
            empty.adjust(slider.x(),0,0,0);
          } else {
            empty.setWidth(slider.x());
            full.adjust(slider.x(),0,0,0);
          }
        else
          if (!opt->upsideDown) {
            full.setHeight(slider.y());
            empty.adjust(0,slider.y(),0,0);
          } else {
            empty.setHeight(slider.y());
            full.adjust(0,slider.y(),0,0);
          }

        if (option->state & State_Enabled) {
          if (option->state & State_MouseOver) {
            renderFrame(painter,empty,fspec,fspec.element+"-focused");
            renderInterior(painter,empty,fspec,ispec,ispec.element+"-focused");
            renderFrame(painter,full,fspec,fspec.element+"-toggled");
            renderInterior(painter,full,fspec,ispec,ispec.element+"-toggled");
          } else {
            renderFrame(painter,empty,fspec,fspec.element+"-normal");
            renderInterior(painter,empty,fspec,ispec,ispec.element+"-normal");
            renderFrame(painter,full,fspec,fspec.element+"-toggled");
            renderInterior(painter,full,fspec,ispec,ispec.element+"-toggled");
          }
        } else {
          renderFrame(painter,empty,fspec,fspec.element+"-disabled");
          renderInterior(painter,empty,fspec,ispec,ispec.element+"-disabled");
          renderFrame(painter,full,fspec,fspec.element+"-disabled");
          renderInterior(painter,full,fspec,ispec,ispec.element+"-disabled");
        }

        group = "PE_SliderCursor";

        fspec = getFrameSpec(group);
        ispec = getInteriorSpec(group);
        indicator_spec_t dspec = getIndicatorSpec(group);

        o.rect = subControlRect(CC_Slider,opt,SC_SliderHandle,widget);

        renderFrame(painter,o.rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,o.rect,fspec,ispec,ispec.element+"-"+status);
        renderIndicator(painter,o.rect,fspec,ispec,dspec,dspec.element+"-"+status);
      }

      break;
    }

    case CC_Dial : { // FIXME
      const QStyleOptionSlider *opt =
          qstyleoption_cast<const QStyleOptionSlider *>(option);

      if (opt) {
        QStyleOptionSlider o(*opt);

        QRect empty(subControlRect(CC_Dial,opt,SC_DialGroove,widget));
        QRect full(subControlRect(CC_Dial,opt,SC_DialHandle,widget));

        renderElement(painter,"dial-empty",empty);
        painter->save();
        painter->setClipRect(full);
        renderElement(painter,"dial-full",empty);
        painter->restore();
      }

      break;
    }

    case CC_TitleBar : {
      const QStyleOptionTitleBar *opt =
        qstyleoption_cast<const QStyleOptionTitleBar *>(option);

      if (opt) {
        QStyleOptionTitleBar o(*opt);

        QString group = "PE_TitleBar";

        frame_spec_t fspec = getFrameSpec(group);
        interior_spec_t ispec = getInteriorSpec(group);
        label_spec_t lspec = getLabelSpec(group);

        renderFrame(painter,o.rect,fspec,fspec.element+"-"+status);
        renderInterior(painter,o.rect,fspec,ispec,ispec.element+"-"+status);

        o.rect = subControlRect(CC_TitleBar,opt,SC_TitleBarLabel,widget);

        renderLabel(painter,o.rect,fspec,ispec,lspec,Qt::AlignLeft | Qt::AlignVCenter, o.text, o.icon.pixmap(pixelMetric(PM_TitleBarHeight)));
      }

      break;
    }

    case CC_MdiControls : {
      printf("mdi\n");

      if (option) {
        QStyleOptionComplex o(*option);

        QString group = "PE_GenericFrame";

        printf("mdi\n");
        frame_spec_t fspec = getFrameSpec(group);

        o.rect = subControlRect(CC_MdiControls,option,SC_MdiCloseButton,widget);
        renderFrame(painter,o.rect,fspec,fspec.element+"-"+status);
      }

      break;
    }

    default : QCommonStyle::drawComplexControl(control,option,painter,widget);
  }

  DEBUG("> drawComplexControl\n");
}

int QuantumStyle::pixelMetric(PixelMetric metric, const QStyleOption * option, const QWidget * widget) const
{
  switch (metric) {
    case PM_ButtonMargin : return 0;

    case PM_ButtonShiftHorizontal :
    case PM_ButtonShiftVertical : return 1;

    case PM_DefaultFrameWidth : {
      QString group = "PE_Toolbar";
      const frame_spec_t fspec = getFrameSpec(group);

      return MAX(MAX(fspec.top,fspec.bottom),MAX(fspec.left,fspec.right));
    }

    case PM_SpinBoxFrameWidth :
    case PM_ComboBoxFrameWidth : return 0;

    case PM_MdiSubWindowFrameWidth : return 2;
    case PM_MdiSubWindowMinimizedWidth : return 50;

//     case PM_LayoutLeftMargin :
//     case PM_LayoutRightMargin :
//     case PM_LayoutTopMargin :
//     case PM_LayoutBottomMargin : return 2;

//     case PM_LayoutHorizontalSpacing :
//     case PM_LayoutVerticalSpacing : return 1;

    case PM_MenuBarPanelWidth :
    case PM_MenuBarVMargin :
    case PM_MenuBarHMargin :  return 0;

    case PM_MenuBarItemSpacing : return 5;

    case PM_MenuTearoffHeight : return 2;

    case PM_MenuPanelWidth : {
      QString group = "PE_MenuItem";
      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      int v = MAX(fspec.top+ispec.top,fspec.bottom+ispec.bottom);
      int h = MAX(fspec.left+ispec.left,fspec.right+ispec.right);
      return MAX(v,h);
    }

    case PM_ToolBarFrameWidth : return 0;
    case PM_ToolBarHandleExtent : return 2;
    case PM_ToolBarSeparatorExtent : return 2;
    case PM_ToolBarItemSpacing : return 0;

    case PM_TabBarTabHSpace : return 0;
    case PM_TabBarTabVSpace : return 0;
    case PM_TabBarBaseOverlap : return 0;
    case PM_TabBarScrollButtonWidth : return 10;

    case PM_ToolBarIconSize : return 16;
    case PM_ToolBarItemMargin : {
      QString group = "PE_Toolbar";
      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      int v = MAX(fspec.top+ispec.top,fspec.bottom+ispec.bottom);
      int h = MAX(fspec.left+ispec.left,fspec.right+ispec.right);
      return MAX(v,h);
    }

    case PM_TabBarIconSize : return 16;
    case PM_SmallIconSize : return 16;
    case PM_LargeIconSize : return 32;

    case PM_FocusFrameHMargin :
    case PM_FocusFrameVMargin : return 0;

    case PM_CheckBoxLabelSpacing :
    case PM_RadioButtonLabelSpacing : return 0;

    case PM_SplitterWidth : return 4;

    case PM_ScrollBarExtent : return 16;
    case PM_ScrollBarSliderMin : return 10;

    case PM_SliderThickness : return 6;

    case PM_ProgressBarChunkWidth : return 20;

    case PM_SliderLength :
    case PM_SliderControlThickness : return 20;

    case PM_DockWidgetFrameWidth : {
      QString group = "PE_Dock";
      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      int v = MAX(fspec.top+ispec.top,fspec.bottom+ispec.bottom);
      int h = MAX(fspec.left+ispec.left,fspec.right+ispec.right);
      return MAX(v,h);
    }

    case PM_DockWidgetTitleMargin : {
      QString group = "PE_Dock";
      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const label_spec_t lspec = getLabelSpec(group);

      int v = MAX(lspec.top,lspec.bottom);
      int h = MAX(lspec.left,lspec.right);
      return MAX(v,h);
    }

    case PM_TextCursorWidth :
      if (widget)
        return QFontMetrics(widget->font()).width("W");
      else
        return 2;

    default : return QCommonStyle::pixelMetric(metric,option,widget);
  }
}

int QuantumStyle::styleHint(StyleHint hint, const QStyleOption * option, const QWidget * widget, QStyleHintReturn * returnData) const
{
  switch (hint) {
    case SH_ComboBox_ListMouseTracking :
    case SH_Menu_MouseTracking :
    case SH_MenuBar_MouseTracking : return true;

    default : return QCommonStyle::styleHint(hint,option,widget,returnData);
  }
}

QCommonStyle::SubControl QuantumStyle::hitTestComplexControl ( ComplexControl control, const QStyleOptionComplex * option, const QPoint & position, const QWidget * widget) const
{
  return QCommonStyle::hitTestComplexControl(control,option,position,widget);
}

QSize QuantumStyle::sizeFromContents ( ContentsType type, const QStyleOption * option, const QSize & contentsSize, const QWidget * widget) const
{
  DEBUG("> sizeFromContents (default)\n");

  int x,y,w,h;
  option->rect.getRect(&x,&y,&w,&h);

  int fh = 14; // font height

  int dw = contentsSize.width();
  int dh = contentsSize.height();

  if (widget)
    fh = QFontMetrics(widget->font()).height();

  QSize s;

  switch (type) {
    case CT_LineEdit : {
      QFont f = QApplication::font();
        if (widget)
          f = widget->font();

      const QString group = "PE_LineEdit";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);
      const label_spec_t lspec = getLabelSpec(group);
      const size_spec_t sspec = getSizeSpec(group);

      s = sizeFromContents(f,fspec,ispec,lspec,sspec,"W",QPixmap());
      s = QSize(s.width() < dw ? dw : s.width(),s.height());
      break;
    }

    case CT_SpinBox : {
      const QStyleOptionSpinBox *opt =
        qstyleoption_cast<const QStyleOptionSpinBox *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_LineEdit";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        if (widget) {
          const QSpinBox *w = qobject_cast<const QSpinBox *>(widget);
          if (w)
            s = sizeFromContents(f,fspec,ispec,lspec,sspec,w->prefix()+QString("%1").arg(w->maximum())+w->suffix(),QPixmap())+QSize(40,0);
        } else
          s = QCommonStyle::sizeFromContents(CT_SpinBox,opt,contentsSize,widget)+QSize(40,0);
      }

      break;
    }

    case CT_ComboBox : {
      const QStyleOptionComboBox *opt =
          qstyleoption_cast<const QStyleOptionComboBox *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_ComboBox";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        QString text = opt->currentText; // longest text

        if (widget) {
          const QComboBox *w = qobject_cast<const QComboBox *>(widget);

          if (w) {
            for (int i=0; i < w->count(); i++)
              if (w->itemText(i).length() > text.length())
                text = w->itemText(i);
          }
        }

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,text,opt->currentIcon.pixmap(opt->iconSize))+QSize(20,0);
      }

      break;
    }

    case CT_PushButton : {
      const QStyleOptionButton *opt =
        qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        QFont f = QApplication::font();

        // set bold for default button
        const QPushButton *w = qobject_cast<const QPushButton *>(widget);
        if (w && w->isDefault()) {
          f = w->font();
          f.setBold(true);
        }

        const QString group = "PE_PanelButtonCommand";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->iconSize));
      }

      break;
    }

    case CT_RadioButton : {
      const QStyleOptionButton *opt =
        qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_RadioButton";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->iconSize))+QSize(pixelMetric(PM_CheckBoxLabelSpacing)+pixelMetric(PM_IndicatorWidth),0);
      }

      break;
    }

    case CT_CheckBox : {
      const QStyleOptionButton *opt =
        qstyleoption_cast<const QStyleOptionButton *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_CheckBox";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->iconSize))+QSize(pixelMetric(PM_CheckBoxLabelSpacing)+pixelMetric(PM_IndicatorWidth),0);
      }

      break;
    }

    case CT_MenuItem : {
      const QStyleOptionMenuItem *opt =
        qstyleoption_cast<const QStyleOptionMenuItem *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_MenuItem";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        if (opt->menuItemType == QStyleOptionMenuItem::Separator)
          s = QSize(dw,pixelMetric(PM_MenuTearoffHeight,option,widget));
        else {
          s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->maxIconWidth));
        }

        if (opt->icon.pixmap(opt->maxIconWidth).isNull())
          s.rwidth() += lspec.tispace+opt->maxIconWidth;

        if ( (opt->menuItemType == QStyleOptionMenuItem::SubMenu) ||
             (opt->checkType == QStyleOptionMenuItem::Exclusive) ||
             (opt->checkType == QStyleOptionMenuItem::NonExclusive)
            ) {
          s.rwidth() += 15+lspec.tispace;
        }
      }

      break;
    }

    case CT_MenuBarItem : {
      const QStyleOptionMenuItem *opt =
        qstyleoption_cast<const QStyleOptionMenuItem *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        QString group = "PE_MenuBarItem";
        frame_spec_t fspec = getFrameSpec(group);
        fspec.left = fspec.right = 0;
        group = "PE_MenuBarItem";
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->maxIconWidth));
      }

      break;
    }

    case CT_ProgressBar : {
      const QStyleOptionProgressBar *opt =
        qstyleoption_cast<const QStyleOptionProgressBar *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_Progressbar";
        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,QPixmap());
      }

      break;
    }

    case CT_ToolButton : {
      const QStyleOptionToolButton *opt =
        qstyleoption_cast<const QStyleOptionToolButton *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_PanelButtonTool";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const indicator_spec_t dspec = getIndicatorSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        if (opt->arrowType == Qt::NoArrow)
          s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->iconSize),opt->toolButtonStyle);
        else
          s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(opt->iconSize),opt->toolButtonStyle)+QSize(lspec.tispace+dspec.size,0);

        if (widget) {
          const QToolButton * w = qobject_cast<const QToolButton *>(widget);
          if (w && w->popupMode() == QToolButton::MenuButtonPopup) {
            s.rwidth() += 20;
          }
        }
      }

      break;
    }

    case CT_TabBarTab : {
      const QStyleOptionTab *opt =
        qstyleoption_cast<const QStyleOptionTab *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_Tab";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(pixelMetric(PM_ToolBarIconSize)));

        if (widget) {
          const QTabBar *w = qobject_cast<const QTabBar*>(widget);
          if (w && w->tabsClosable())
            s.rwidth() += pixelMetric(PM_TabCloseIndicatorWidth,option,widget)+lspec.tispace;
        }
      }

      break;
    }

    case CT_HeaderSection : {
      const QStyleOptionHeader *opt =
        qstyleoption_cast<const QStyleOptionHeader *>(option);

      if (opt) {
        QFont f = QApplication::font();
        if (widget)
          f = widget->font();

        const QString group = "PE_HeaderSection";

        const frame_spec_t fspec = getFrameSpec(group);
        const interior_spec_t ispec = getInteriorSpec(group);
        const label_spec_t lspec = getLabelSpec(group);
        const size_spec_t sspec = getSizeSpec(group);

        s = sizeFromContents(f,fspec,ispec,lspec,sspec,opt->text,opt->icon.pixmap(pixelMetric(PM_SmallIconSize)));
      }

      break;
    }

    case CT_Slider : {
      if (option->state & State_Horizontal)
        s = QSize(dw,pixelMetric(PM_SliderControlThickness,option,widget)+2); // +2 for frame
      else
        s = QSize(pixelMetric(PM_SliderControlThickness,option,widget),dh+2);

      break;
    }

    default : s = QCommonStyle::sizeFromContents(type,option,contentsSize,widget);
  }

  DEBUG("< sizeFromContents (default)\n");
  return s;
}

QSize QuantumStyle::sizeFromContents(const QFont &font,
                     /* frame spec */ const frame_spec_t &fspec,
                     /* interior spec */ const interior_spec_t &ispec,
                     /* label spec */ const label_spec_t &lspec,
                     /* size spec */ const size_spec_t &sspec,
                     /* text */ const QString &text,
                     /* icon */ const QPixmap &icon,
                     /* text-icon alignment */ const Qt::ToolButtonStyle tialign) const
{
  DEBUG("> sizeFromContents (custom)\n");

  QSize s;
  s.setWidth(fspec.left+fspec.right+ispec.left+ispec.right+lspec.left+lspec.right);
  s.setHeight(fspec.top+fspec.bottom+ispec.top+ispec.bottom+lspec.top+lspec.bottom);
  if (lspec.hasShadow) {
    s.rwidth() += lspec.xshift+lspec.depth;
    s.rheight() += lspec.yshift+lspec.depth;
  }

  // compute width and height of text. QFontMetrics seems to ignore \n
  // and returns wrong values

  int tw = 0;
  int th = 0;

  /* remove & mnemonic character and tabs (for menu items) */
  QString t = QString(text).remove('&').remove('\t');

  /* deal with \n */
  QStringList l = t.split('\n');
  th = QFontMetrics(font).height()*(l.size());
  for (int i=0; i<l.size(); i++) {
    tw = MAX(tw,QFontMetrics(font).width(l[i]));
  }

  if (tialign == Qt::ToolButtonIconOnly) {
    s.rwidth() += icon.width();
    s.rheight() += icon.height();
  } else if (tialign == Qt::ToolButtonTextOnly) {
    s.rwidth() += tw;
    s.rheight() += th;
  } else if (tialign == Qt::ToolButtonTextBesideIcon) {
    s.rwidth() += icon.width() + (icon.isNull() ? 0 : lspec.tispace) + tw;
    s.rheight() += MAX(icon.height(),th);
  } else if (tialign == Qt::ToolButtonTextUnderIcon) {
    s.rwidth() += MAX(icon.width(),tw);
    s.rheight() += icon.height() + (icon.isNull() ? 0 : lspec.tispace) + th;
  }

  if ( (sspec.minH > 0) && (s.height() < sspec.minH) )
    s.setHeight(sspec.minH);

  if ( (sspec.minW > 0) && (s.width() < sspec.minW) )
    s.setWidth(sspec.minW);

  if (sspec.fixedH > 0)
    s.setHeight(sspec.fixedH);

  if (sspec.fixedW > 0)
    s.setWidth(sspec.fixedW);

  DEBUG("< sizeFromContents (custom)\n");
  return s;
}

QRect QuantumStyle::subElementRect(SubElement element, const QStyleOption * option, const QWidget * widget) const
{
  switch (element) {
    case SE_CheckBoxFocusRect :
    case SE_RadioButtonFocusRect :
    case SE_ProgressBarGroove :
    case SE_HeaderLabel :
    case SE_ProgressBarLabel : return option->rect;
    case SE_ProgressBarContents : {
      const QString group = "PE_Progressbar";

      const frame_spec_t fspec = getFrameSpec(group);
      const interior_spec_t ispec = getInteriorSpec(group);

      return interiorRect(option->rect, fspec,ispec);
    }

    default : return QCommonStyle::subElementRect(element,option,widget);
  }
}

QRect QuantumStyle::subControlRect(ComplexControl control, const QStyleOptionComplex * option, SubControl subControl, const QWidget * widget) const
{
  int x,y,h,w;
  option->rect.getRect(&x,&y,&w,&h);

  switch (control) {
    case CC_TitleBar :
      switch (subControl) {
        case SC_TitleBarLabel : return option->rect;

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }
      break;

    case CC_SpinBox :
      switch (subControl) {
        case SC_SpinBoxFrame : return QRect();
        case SC_SpinBoxEditField : return QRect(x,y,w-40,h);
        case SC_SpinBoxUp : return QRect(x+w-20,y,20,h);
        case SC_SpinBoxDown : return QRect(x+w-40,y,20,h);

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }
      break;

    case CC_ComboBox :

      switch (subControl) {
        case SC_ComboBoxFrame : return QRect();
        case SC_ComboBoxEditField : return QRect(x,y,w-20,h);
        case SC_ComboBoxArrow : return QRect(x+w-20,y,20,h);
        case SC_ComboBoxListBoxPopup : {
          const QString group = "PE_ComboBox";

          const frame_spec_t fspec = getFrameSpec(group);
          const interior_spec_t ispec = getInteriorSpec(group);
          const label_spec_t lspec = getLabelSpec(group);
          const size_spec_t sspec = getSizeSpec(group);

          int maxW = 0;

          if (widget) {
            const QComboBox * w = qobject_cast<const QComboBox *>(widget);
            QSize s;
            for (int i=0; i<w->count(); i++) {
              s = sizeFromContents(w->font(),fspec,ispec,lspec,sspec, w->itemText(i),w->itemIcon(i).pixmap(w->iconSize()));
              if (s.width() > maxW)
                maxW = s.width();
            }

            if (w->count() > w->maxVisibleItems()) {
              /* add scrollbar width */
              maxW = maxW+pixelMetric(PM_ScrollBarExtent);
            }
          }

          if (maxW < w)
            maxW = w; /* popup width not shorter than combo itself */

          return QRect(0,h,maxW,0 /* height seems not to be used */);
        }

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }
      break;

    case CC_MdiControls :
      switch (subControl) {
        case SC_MdiCloseButton : return option->rect;

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }
      break;

    case CC_ScrollBar : {
      const int extent = pixelMetric(PM_ScrollBarExtent,option,widget);
      if (option->state & State_Horizontal)
        switch (subControl) {
          case SC_ScrollBarGroove : return QRect(x+extent,y,w-2*extent,h);
          case SC_ScrollBarSubLine : return QRect(x,y,extent,extent);
          case SC_ScrollBarAddLine : return QRect(x+w-extent,y,extent,extent);
          case SC_ScrollBarSlider : {
            const QStyleOptionSlider *opt =
                qstyleoption_cast<const QStyleOptionSlider *>(option);

            if (opt) {
              QRect r = subControlRect(CC_ScrollBar,option,SC_ScrollBarGroove,widget);
              r.getRect(&x,&y,&w,&h);

              const int minLength = pixelMetric(PM_ScrollBarSliderMin,option,widget);
              const int maxLength = w; // max slider length
              const int valueRange = opt->maximum - opt->minimum;
              int length = maxLength;
              if (opt->minimum != opt->maximum) {
                length = (opt->pageStep*maxLength) / (valueRange+opt->pageStep);

                if ( (length < minLength) || (valueRange > INT_MAX/2) )
                  length = minLength;
                if (length > maxLength)
                  length = maxLength;
              }

              const int start = sliderPositionFromValue(opt->minimum,opt->maximum,opt->sliderPosition,maxLength - length,opt->upsideDown);
              return QRect(x+start,y,length,h);
            } else
              return QRect();
          }

          default : return QCommonStyle::subControlRect(control,option,subControl,widget);

          break;
        }
      else
        switch (subControl) {
          case SC_ScrollBarGroove : return QRect(x,y+extent,w,h-2*extent);
          case SC_ScrollBarSubLine : return QRect(x,y,extent,extent);
          case SC_ScrollBarAddLine : return QRect(x,y+h-extent,extent,extent);
          case SC_ScrollBarSlider : {
            const QStyleOptionSlider *opt =
                qstyleoption_cast<const QStyleOptionSlider *>(option);

            if (opt) {
              QRect r = subControlRect(CC_ScrollBar,option,SC_ScrollBarGroove,widget);
              r.getRect(&x,&y,&w,&h);

              const int minLength = pixelMetric(PM_ScrollBarSliderMin,option,widget);
              const int maxLength = h; // max slider length
              const int valueRange = opt->maximum - opt->minimum;
              int length = maxLength;
              if (opt->minimum != opt->maximum) {
                length = (opt->pageStep*maxLength) / (valueRange+opt->pageStep);

                if ( (length < minLength) || (valueRange > INT_MAX/2) )
                  length = minLength;
                if (length > maxLength)
                  length = maxLength;
              }

              const int start = sliderPositionFromValue(opt->minimum,opt->maximum,opt->sliderPosition,maxLength - length,opt->upsideDown);
              return QRect(x,y+start,w,length);
            } else
              return QRect();
          }

          default : return QCommonStyle::subControlRect(control,option,subControl,widget);

          break;
        }

      break;
    }

    case CC_Slider : {
      const int thick = pixelMetric(PM_SliderThickness,option,widget);
      const bool horiz = (option->state & State_Horizontal);
      switch (subControl) {
        case SC_SliderGroove :
          if (horiz)
            return QRect(x+1,y+(h-thick)/2+1,w-2,thick);
          else
            return QRect(x+(w-thick)/2+1,y+1,thick,h-2);

        case SC_SliderHandle : {
          const QStyleOptionSlider *opt =
            qstyleoption_cast<const QStyleOptionSlider *>(option);

          if (opt) {
            subControlRect(CC_Slider,option,SC_SliderGroove,widget).getRect(&x,&y,&w,&h);

            const int len = pixelMetric(PM_SliderLength, option, widget);
            const int thickness = pixelMetric(PM_SliderControlThickness, option, widget);
            const int sliderPos(sliderPositionFromValue(opt->minimum, opt->maximum, opt->sliderPosition, (horiz ? w : h) - len, opt->upsideDown));

            if (horiz)
              return QRect(x+sliderPos,y+(h-thickness)/2,len,thickness);
            else
              return QRect(x+(w-len)/2,y+sliderPos,thickness,len);
          }

        }

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }

      break;
    }

    case CC_Dial : {
      const bool horiz = (option->state & State_Horizontal);
      switch(subControl) {
        case SC_DialGroove : return option->rect;
        case SC_DialHandle : {
          const QStyleOptionSlider *opt =
            qstyleoption_cast<const QStyleOptionSlider *>(option);

          if (opt) {
            subControlRect(CC_Dial,option,SC_DialGroove,widget).getRect(&x,&y,&w,&h);

            const int sliderPos(sliderPositionFromValue(opt->minimum, opt->maximum, opt->sliderPosition, (horiz ? w : h), opt->upsideDown));

            if (horiz)
              return QRect(x,y,w,h-sliderPos);
            else
              return QRect(x,y,w-sliderPos,h);
          }

        }

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }

      break;
    }

    case CC_ToolButton : {
      switch (subControl) {
        case SC_ToolButton : {
          const QStyleOptionToolButton *opt =
            qstyleoption_cast<const QStyleOptionToolButton *>(option);

          if (opt) {
            if (widget) {
              const QToolButton * w = qobject_cast<const QToolButton *>(widget);
              if (w && w->popupMode() == QToolButton::MenuButtonPopup)
                return option->rect.adjusted(0,0,-20,0);
              else
                return option->rect;
            } else
              return option->rect;
          }
        }

        case SC_ToolButtonMenu : {
            return QRect(x+w-20,y,20,h);
        }

        default : return QCommonStyle::subControlRect(control,option,subControl,widget);
      }

      break;
    }

    default : return QCommonStyle::subControlRect(control,option,subControl,widget);
  }

  return QCommonStyle::subControlRect(control,option,subControl,widget);
}

void QuantumStyle::polish(QWidget * widget)
{
  if (widget)
    widget->setAttribute(Qt::WA_Hover, true);
}

void QuantumStyle::unpolish(QWidget * widget)
{
  if (widget)
    widget->setAttribute(Qt::WA_Hover, false);
}

QRect QuantumStyle::squaredRect(const QRect& r) const {
  int e = (r.width() > r.height()) ? r.height() : r.width();
  return QRect(r.x(),r.y(),e,e);
}

void QuantumStyle::renderElement(QPainter * painter, const QString & element, const QRect & bounds, bool pattern, int hsize,int vsize, Qt::Orientation orientation) const
{
  int x,y,h,w;
  bounds.getRect(&x,&y,&w,&h);

  QSvgRenderer *renderer = 0;

  // Render application specific widget
  if (appRndr->isValid() && appRndr->elementExists(element))
      renderer = appRndr;
  else
    // Fall back to theme specific rendering
    if (themeRndr->isValid() && themeRndr->elementExists(element))
        renderer = themeRndr;
    else
      // Fall back to default rendering
      if (defaultRndr->isValid() && defaultRndr->elementExists(element))
        renderer = defaultRndr;

  if (renderer) {
    if (!pattern)
      renderer->render(painter,element,QRect(x,y,w,h));
    else if ( (hsize > 0) || (vsize > 0) ) {

      if ( (hsize > 0) && (vsize <= 0) ) {
        int hpatterns = (w/hsize)+1;

        painter->save();
        painter->setClipRect(QRect(x,y,w,h));
        for (int i=0; i<hpatterns; i++)
          renderer->render(painter,element,QRect(x+i*hsize,y,hsize,h));
        painter->restore();
      }

      if ( (hsize <= 0) && (vsize > 0) ) {
        int vpatterns = (h/vsize)+1;

        painter->save();
        painter->setClipRect(QRect(x,y,w,h));
        for (int i=0; i<vpatterns; i++)
          renderer->render(painter,element,QRect(x,y+i*vsize,w,vsize));
        painter->restore();
      }

      if ( (hsize > 0) && (vsize > 0) ) {
        int hpatterns = (w/hsize)+1;
        int vpatterns = (h/vsize)+1;

        painter->save();
        painter->setClipRect(bounds);
        for (int i=0; i<hpatterns; i++)
          for (int j=0; j<vpatterns; j++)
            renderer->render(painter,element,QRect(x+i*hsize,y+j*vsize,hsize,vsize));
        painter->restore();
      }
    }
  }
}

void QuantumStyle::renderFrame(QPainter *painter,
                    /* frame bounds */ const QRect &bounds,
                    /* frame spec */ const frame_spec_t &fspec,
                    /* SVG element */ const QString &element) const
{
  DEBUG("BEGIN renderFrame %s\n",(const char *)(element.toAscii()));

  if (!fspec.hasFrame)
    return;

  int x0,y0,x1,y1,w,h;
  bounds.getRect(&x0,&y0,&w,&h);
  x1 = bounds.bottomRight().x();
  y1 = bounds.bottomRight().y();

  if (!fspec.hasCapsule) {
    // top
    renderElement(painter,element+"-top",QRect(x0+fspec.left,y0,w-fspec.left-fspec.right,fspec.top),fspec.rtop,fspec.ptop,0);
    // bottom
    renderElement(painter,element+"-bottom",QRect(x0+fspec.left,y1-fspec.bottom+1,w-fspec.left-fspec.right,fspec.bottom),fspec.rbottom,fspec.pbottom,0);
    // left
    renderElement(painter,element+"-left",QRect(x0,y0+fspec.top,fspec.left,h-fspec.top-fspec.bottom),fspec.rleft,0,fspec.pleft);
    // right
    renderElement(painter,element+"-right",QRect(x1-fspec.right+1,y0+fspec.top,fspec.right,h-fspec.top-fspec.bottom),fspec.rright,0,fspec.pright);
    // topleft
    renderElement(painter,element+"-topleft",QRect(x0,y0,fspec.left,fspec.top));
    // topright
    renderElement(painter,element+"-topright",QRect(x1-fspec.right+1,y0,fspec.right,fspec.top));
    // bottomleft
    renderElement(painter,element+"-bottomleft",QRect(x0,y1-fspec.bottom+1,fspec.left,fspec.bottom));
    // bottomright
    renderElement(painter,element+"-bottomright",QRect(x1-fspec.right+1,y1-fspec.bottom+1,fspec.right,fspec.bottom));
  } else {
    // edges
    // top
    if ( (fspec.capsuleV == -1) || (fspec.capsuleV == 2) )
      renderElement(painter,element+"-top",QRect(x0+fspec.left,y0,w-fspec.left-fspec.right,fspec.top),fspec.rtop,fspec.ptop,0);
    // bottom
    if ( (fspec.capsuleV == 1) || (fspec.capsuleV == 2) )
      renderElement(painter,element+"-bottom",QRect(x0+fspec.left,y1-fspec.bottom+1,w-fspec.left-fspec.right,fspec.bottom),fspec.rbottom,fspec.pbottom,0);
    // left
    if ( (fspec.capsuleH == -1) || (fspec.capsuleH == 2) )
      renderElement(painter,element+"-left",QRect(x0,y0+fspec.top,fspec.left,h-fspec.top-fspec.bottom),fspec.rleft,0,fspec.pleft);
    // right
    if ( (fspec.capsuleH == 1) || (fspec.capsuleH == 2) )
      renderElement(painter,element+"-right",QRect(x1-fspec.right+1,y0+fspec.top,fspec.right,h-fspec.top-fspec.bottom),fspec.rright,0,fspec.pright);

    // corners : top row
    if ( (fspec.capsuleV == -1) || (fspec.capsuleV == 2) ) {
      // topleft
      if ( (fspec.capsuleH == 0) || (fspec.capsuleH == 1) )
        renderElement(painter,element+"-top",QRect(x0,y0,fspec.left,fspec.top),fspec.rtop,fspec.ptop,0);
      if (fspec.capsuleH == -1)
        renderElement(painter,element+"-topleft",QRect(x0,y0,fspec.left,fspec.top));

      // topright
      if ( (fspec.capsuleH == -1) || (fspec.capsuleH == 0) )
        renderElement(painter,element+"-top",QRect(x1-fspec.right+1,y0,fspec.right,fspec.top),fspec.rtop,fspec.ptop,0);
      if (fspec.capsuleH == 1)
        renderElement(painter,element+"-topright",QRect(x1-fspec.right+1,y0,fspec.right,fspec.top));
    }

    // corners : bottom row
    if ( (fspec.capsuleV == 1) || (fspec.capsuleV == 2) ) {
      // bottomleft
      if ( (fspec.capsuleH == 0) || (fspec.capsuleH == 1) )
        renderElement(painter,element+"-bottom",QRect(x0,y1-fspec.bottom+1,fspec.left,fspec.bottom),fspec.rbottom,fspec.pbottom,0);
      if (fspec.capsuleH == -1)
        renderElement(painter,element+"-bottomleft",QRect(x0,y1-fspec.bottom+1,fspec.left,fspec.bottom));

      // bottomright
      if ( (fspec.capsuleH == -1) || (fspec.capsuleH == 0) )
        renderElement(painter,element+"-bottom",QRect(x1-fspec.right+1,y1-fspec.bottom+1,fspec.right,fspec.bottom),fspec.rbottom,fspec.pbottom,0);
      if (fspec.capsuleH == 1)
        renderElement(painter,element+"-bottomright",QRect(x1-fspec.right+1,y1-fspec.bottom+1,fspec.right,fspec.bottom));
    }

    // corners : left column
    if ( (fspec.capsuleH == -1) || (fspec.capsuleH == 2) ) {
      // topleft
      if ( (fspec.capsuleV == 0) || (fspec.capsuleV == 1) )
        renderElement(painter,element+"-left",QRect(x0,y0,fspec.left,fspec.top),fspec.rleft,0,fspec.pleft);;
      if (fspec.capsuleV == -1)
        renderElement(painter,element+"-topleft",QRect(x0,y0,fspec.left,fspec.top));

      // bottomleft
      if ( (fspec.capsuleV == -1) || (fspec.capsuleV == 0) )
        renderElement(painter,element+"-left",QRect(x0,y1-fspec.bottom+1,fspec.left,fspec.bottom),fspec.rleft,0,fspec.pleft);;
      if (fspec.capsuleV == 1)
        renderElement(painter,element+"-bottomleft",QRect(x0,y1-fspec.bottom+1,fspec.left,fspec.bottom));
    }

    // corners : right column
    if ( (fspec.capsuleH == 1) || (fspec.capsuleH == 2) ) {
      // topleft
      if ( (fspec.capsuleV == 0) || (fspec.capsuleV == 1) )
        renderElement(painter,element+"-right",QRect(x1-fspec.right+1,y0,fspec.right,fspec.top),fspec.rright,0,fspec.pright);;
      if (fspec.capsuleV == -1)
        renderElement(painter,element+"-topright",QRect(x1-fspec.right+1,y0,fspec.right,fspec.top));

      // bottomleft
      if ( (fspec.capsuleV == -1) || (fspec.capsuleV == 0) )
        renderElement(painter,element+"-right",QRect(x1-fspec.right+1,y1-fspec.bottom+1,fspec.right,fspec.bottom),fspec.rright,0,fspec.pright);;
      if (fspec.capsuleV == 1)
        renderElement(painter,element+"-bottomright",QRect(x1-fspec.right+1,y1-fspec.bottom+1,fspec.right,fspec.bottom));
    }
  }

  #ifdef __DEBUG__
  painter->save();
  painter->setPen(QPen(Qt::green));
  drawRealRect(painter, bounds);
  painter->restore();
  #endif

  DEBUG("END renderFrame %s\n",(const char *)(element.toAscii()));
}

void QuantumStyle::renderInterior(QPainter *painter,
                       /* frame bounds */ const QRect &bounds,
                       /* frame spec */ const frame_spec_t &fspec,
                       /* interior spec */ const interior_spec_t &ispec,
                       /* SVG element */ const QString &element) const
{
  DEBUG("BEGIN renderInterior %s\n",(const char *)(element.toAscii()));

  if (!ispec.hasInterior)
    return;

  if (!fspec.hasCapsule)
    renderElement(painter,element,interiorRect(bounds,fspec,ispec),(ispec.px || ispec.py),ispec.px,ispec.py);
  else {
    // add these to compensate the absence of the frame
    int left = 0, right = 0, top = 0, bottom = 0;
    if (fspec.capsuleH == 0) {
      left = fspec.left+ispec.left;
      right = fspec.right+ispec.right;
    }
    if (fspec.capsuleH == -1) {
      right = fspec.right+ispec.right;
    }
    if (fspec.capsuleH == 1) {
      left = fspec.left+ispec.left;
    }
    if (fspec.capsuleV == 0) {
      top = fspec.top+ispec.top;
      bottom = fspec.bottom+ispec.bottom;
    }
    if (fspec.capsuleV == -1) {
      bottom = fspec.bottom+ispec.bottom;
    }
    if (fspec.capsuleV == 1) {
      top = fspec.top+ispec.top;
    }
    renderElement(painter,element,
                   interiorRect(bounds,fspec,ispec).adjusted(-left,-top,right,bottom),
                   (ispec.px || ispec.py),ispec.px,ispec.py);
  }

  #ifdef __DEBUG__
  painter->save();
  painter->setPen(QPen(Qt::red));
  drawRealRect(painter,interiorRect(bounds,fspec,ispec));
  painter->restore();
  #endif

  DEBUG("END renderInterior %s\n",(const char *)(element.toAscii()));
}

void QuantumStyle::renderIndicator(QPainter *painter,
                       /* frame bounds */ const QRect &bounds,
                       /* frame spec */ const frame_spec_t &fspec,
                       /* interior spec */ const interior_spec_t &ispec,
                       /* indicator spec */ const indicator_spec_t &dspec,
                       /* indocator SVG element */ const QString &element,
                       Qt::Alignment alignment) const
{
  DEBUG("BEGIN renderIndicator %s\n",(const char *)(element.toAscii()));

  const QRect interior = squaredRect(interiorRect(bounds,fspec,ispec));
  //const QRect interior = squaredRect(frameRect(bounds,fspec));
  int s = (interior.width() > dspec.size) ? dspec.size : interior.width();

  renderElement(painter,element,
                 alignedRect(QApplication::layoutDirection(),alignment,QSize(s,s),bounds));

  #ifdef __DEBUG__
  painter->save();
  painter->setPen(QPen(Qt::blue));
  drawRealRect(painter, alignedRect(QApplication::layoutDirection(),alignment,QSize(s,s),bounds));
  painter->restore();
  #endif

  DEBUG("END renderIndicator %s\n",(const char *)(element.toAscii()));
}

void QuantumStyle::renderLabel(QPainter *painter,
                     /* frame bounds */ const QRect &bounds,
                     /* frame spec */ const frame_spec_t &fspec,
                     /* interior spec */ const interior_spec_t &ispec,
                     /* label spec */ const label_spec_t &lspec,
                     /* text alignment */ int talign,
                     /* text */ const QString &text,
                     /* icon */ const QPixmap &icon,
                     /* text-icon alignment */ const Qt::ToolButtonStyle tialign) const
{
  DEBUG("BEGIN renderLabel %s\n",(const char *)(text.toAscii()));

  // compute text and icon rect
  QRect r(textRect(bounds,fspec,ispec,lspec));

  #ifdef __DEBUG__
  painter->save();
  painter->setPen(QPen(QColor(255,255,255)));
  drawRealRect(painter, r);
  painter->restore();
  #endif

  QRect ricon = r;
  QRect rtext = r;

  if (tialign == Qt::ToolButtonTextBesideIcon) {
    ricon = alignedRect(QApplication::layoutDirection(),Qt::AlignVCenter | Qt::AlignLeft, QSize(icon.width(),icon.height()),r);
    rtext = QRect(r.x()+icon.width()+(icon.isNull() ? 0 : lspec.tispace),r.y(),r.width()-ricon.width()-(icon.isNull() ? 0 : lspec.tispace),r.height());
  } else if (tialign == Qt::ToolButtonTextUnderIcon) {
    ricon = alignedRect(QApplication::layoutDirection(),Qt::AlignTop | Qt::AlignHCenter, QSize(icon.width(),icon.height()),r);
    rtext = QRect(r.x(),r.y()+icon.height()+(icon.isNull() ? 0 : lspec.tispace),r.width(),r.height()-ricon.height()-(icon.isNull() ? 0 : lspec.tispace));
  } else if (tialign == Qt::ToolButtonIconOnly) {
    ricon = alignedRect(QApplication::layoutDirection(),Qt::AlignCenter, QSize(icon.width(),icon.height()),r);
  }

  if (tialign != Qt::ToolButtonIconOnly) {
    if (lspec.hasShadow) {
      painter->save();

      painter->setPen(QPen(QColor(lspec.r,lspec.g,lspec.b,lspec.a)));
      for (int i=0; i<lspec.depth; i++)
        painter->drawText(rtext.adjusted(lspec.xshift+i,lspec.yshift+i,0,0),talign,text);

      painter->restore();
    }
    painter->drawText(rtext,talign,text);
  }

  if (tialign != Qt::ToolButtonTextOnly) {
    if (!icon.isNull()) {
      painter->drawPixmap(ricon,icon);
#ifdef __DEBUG__
      painter->save();
      painter->setPen(QPen(QColor(255,0,255)));
      drawRealRect(painter, ricon);
      painter->restore();
#endif
    }
  }

#ifdef __DEBUG__
  painter->save();
  painter->setPen(QPen(QColor(0,255,255)));
  drawRealRect(painter, rtext);
  painter->restore();
#endif

  DEBUG("END renderLabel %s\n",(const char *)(text.toAscii()));
}

QVariant QuantumStyle::getValue(const QString &group, const QString &key, bool *ok) const
{
  if (appSettings && appSettings->isValid())
    return appSettings->getValue(group,key,ok);
  else if (themeSettings && themeSettings->isValid())
    return themeSettings->getValue(group,key,ok);
  else
    return defaultSettings->getValue(group,key,ok);
}

frame_spec_t QuantumStyle::getFrameSpec(const QString &widgetName) const
{
  frame_spec_t r;
  r.hasFrame = false;
  r.hasCapsule = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.rtop = r.rbottom = r.rleft = r.rright = false;
  r.ptop = r.pbottom = r.pleft = r.pright = 0;
  r.capsuleH = r.capsuleV = 0;

  bool ok;

  QVariant v = getValue(widgetName,"frame", &ok);
  if (ok) {
    r.hasFrame = v.toBool();
    if (r.hasFrame) {
      v = getValue(widgetName, "frame.element", &ok);
      if (ok)
        r.element = v.toString();
      v = getValue(widgetName,"frame.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"frame.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"frame.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"frame.right", &ok);
      if (ok)
        r.right = v.toInt();

      v = getValue(widgetName,"frame.capsule", &ok);
      if (ok)
        r.hasCapsule = v.toBool();

      v = getValue(widgetName,"frame.repeat.top",&ok);
      if (ok)
        r.rtop = v.toBool();
      v = getValue(widgetName,"frame.repeat.bottom",&ok);
      if (ok)
        r.rbottom = v.toBool();
      v = getValue(widgetName,"frame.repeat.left",&ok);
      if (ok)
        r.rleft = v.toBool();
      v = getValue(widgetName,"frame.repeat.right",&ok);
      if (ok)
        r.rright = v.toBool();

      if (r.rtop) {
        v = getValue(widgetName,"frame.repeat.top.patternsize",&ok);
        if (ok)
          r.ptop = v.toInt();
      }
      if (r.rbottom) {
        v = getValue(widgetName,"frame.repeat.bottom.patternsize",&ok);
        if (ok)
          r.pbottom = v.toInt();
      }
      if (r.rleft) {
        v = getValue(widgetName,"frame.repeat.left.patternsize",&ok);
        if (ok)
          r.pleft = v.toInt();
      }
      if (r.rright) {
        v = getValue(widgetName,"frame.repeat.right.patternsize",&ok);
        if (ok)
          r.pright = v.toInt();
      }
    }
  }

  return r;
}

interior_spec_t QuantumStyle::getInteriorSpec(const QString &widgetName) const
{
  interior_spec_t r;
  r.hasInterior = true;
  r.hasMargin = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.rx = r.ry = false;
  r.px = r.py = 0;

  bool ok;

  QVariant v = getValue(widgetName,"interior", &ok);
  if (ok) {
    r.hasInterior = v.toBool();
  }
  if (r.hasInterior) {
    v = getValue(widgetName, "interior.element", &ok);
      if (ok)
        r.element = v.toString();
  }

  v = getValue(widgetName,"interior.margin", &ok);
  if (ok) {
    r.hasMargin = v.toBool();
    if (r.hasMargin) {
      v = getValue(widgetName,"interior.margin.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"interior.margin.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"interior.margin.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"interior.margin.right", &ok);
      if (ok)
        r.right = v.toInt();
    }
  }

  v = getValue(widgetName,"interior.repeat.x",&ok);
  if (ok)
    r.rx = v.toBool();
  v = getValue(widgetName,"interior.repeat.y",&ok);
  if (ok)
    r.ry = v.toBool();

  if (r.rx) {
    v = getValue(widgetName,"interior.repeat.x.patternsize",&ok);
    if (ok)
      r.px = v.toInt();
  }

  if (r.ry) {
    v = getValue(widgetName,"interior.repeat.y.patternsize",&ok);
    if (ok)
      r.py = v.toInt();
  }

  return r;
}

indicator_spec_t QuantumStyle::getIndicatorSpec(const QString &widgetName) const
{
  indicator_spec_t r;
  r.size = 15;

  bool ok;

  QVariant v = getValue(widgetName,"indicator.size", &ok);
  if (ok)
    r.size = v.toInt();

  v = getValue(widgetName, "indicator.element", &ok);
    if (ok)
      r.element = v.toString();

  return r;
}

label_spec_t QuantumStyle::getLabelSpec(const QString &widgetName) const
{
  label_spec_t r;
  r.hasShadow = false;
  r.xshift = 0;
  r.yshift = 1;
  r.r = r.g = r.b = 0;
  r.a = 255;
  r.depth = 1;
  r.hasMargin = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.tispace = 0;

  bool ok;

  QVariant v = getValue(widgetName,"text.shadow", &ok);
  if (ok) {
    r.hasShadow = v.toBool();

    if (r.hasShadow) {
      v = getValue(widgetName,"text.shadow.xshift", &ok);
      if (ok)
        r.xshift = v.toInt();
      v = getValue(widgetName,"text.shadow.yshift", &ok);
      if (ok)
        r.yshift = v.toInt();
      v = getValue(widgetName,"text.shadow.red", &ok);
      if (ok)
        r.r = v.toInt();
      v = getValue(widgetName,"text.shadow.green", &ok);
      if (ok)
        r.g = v.toInt();
      v = getValue(widgetName,"text.shadow.blue", &ok);
      if (ok)
        r.b = v.toInt();
      v = getValue(widgetName,"text.shadow.alpha", &ok);
      if (ok)
        r.a = v.toInt();
      v = getValue(widgetName,"text.shadow.depth", &ok);
      if (ok)
        r.depth = v.toInt();
    }
  }

  v = getValue(widgetName,"text.margin", &ok);
  if (ok) {
    r.hasMargin = v.toBool();
    if (r.hasMargin) {
      v = getValue(widgetName,"text.margin.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"text.margin.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"text.margin.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"text.margin.right", &ok);
      if (ok)
        r.right = v.toInt();
    }
  }

  v = getValue(widgetName,"text.iconspacing", &ok);
  if (ok)
    r.tispace = v.toInt();

  return r;
}

size_spec_t QuantumStyle::getSizeSpec(const QString& widgetName) const
{
  size_spec_t r;
  r.minH = r.fixedH = r.minW = r.fixedW = -1;

  bool ok;

  QVariant v = getValue(widgetName,"size.minheight", &ok);
  if (ok)
    r.minH = v.toInt();
  v = getValue(widgetName,"size.fixedheight", &ok);
  if (ok)
    r.fixedH = v.toInt();

  v = getValue(widgetName,"size.minwidth", &ok);
  if (ok)
    r.minW = v.toInt();
  v = getValue(widgetName,"size.fixedwidth", &ok);
  if (ok)
    r.fixedW = v.toInt();

  return r;
}

void QuantumStyle::capsulePosition(const QWidget *widget, bool &capsule, int &h, int &v) const
{
  capsule = false;
  h = v = 2;
  if (widget && widget->parent()) {
    // get parent widget
    const QWidget *parent = qobject_cast<const QWidget *>(widget->parent());
    if (parent) {
      // get its layout
      QLayout *l = parent->layout();
      if (l) {
        capsule = true;
        // ensure that all widgets in the layout have the same type
        for (int i = 0; i < l->count(); ++i) {
          if (l->itemAt(i)->widget()) {
            //printf("classname %s\n",l->itemAt(i)->widget()->metaObject()->className());
            if (l->itemAt(i)->widget()->metaObject()->className() != widget->metaObject()->className()) {
              capsule = false;
              break;
            }
          }
        }

        if (capsule) {
          int index = -1;
          // get index of this widget
          for (int i = 0; i < l->count(); ++i) {
            if (l->itemAt(i)->widget() == widget) {
              index = i;
              break;
            }
          }

          const QHBoxLayout *hbox = qobject_cast<const QHBoxLayout *>(l);
          if (hbox) {
            // layout is a horizontal box
            if ( (index == 0) && (index == hbox->count()-1) )
              h = 2;
            else if (index == hbox->count()-1)
              h = 1;
            else if (index == 0)
              h = -1;
            else
              h = 0;

            v = 2;
          }

          const QVBoxLayout *vbox = qobject_cast<const QVBoxLayout *>(l);
          if (vbox) {
            // layout is a horizontal box
            if ( (index == 0) && (index == vbox->count()-1) )
              v = 2;
            else if (index == vbox->count()-1)
              v = 1;
            else if (index == 0)
              v = -1;
            else
              v = 0;

            h = 2;
          }

          const QGridLayout *gbox = qobject_cast<const QGridLayout *>(l);
          if (gbox) {
            // layout is a grid
            const int rows = gbox->rowCount();
            const int cols = gbox->columnCount();

            if (rows == 1)
              v = 2;
            else if (index < cols)
              v = -1;
            else if (index > cols*(rows-1)-1)
              v = 1;
            else
              v = 0;

            if (cols == 1)
              h = 2;
            else if (index%cols == 0)
              h = -1;
            else if (index%cols == cols-1)
              h = 1;
            else
              h = 0;
          }
        }
      }
    }
  }
}

void QuantumStyle::drawRealRect(QPainter* p, const QRect& r) const
{
  int x0,y0,x1,y1,w,h;
  r.getRect(&x0,&y0,&w,&h);
  x1 = r.bottomRight().x();
  y1 = r.bottomRight().y();
  p->drawLine(x0,y0,x1,y0);
  p->drawLine(x0,y0,x0,y1);
  p->drawLine(x1,y0,x1,y1);
  p->drawLine(x0,y1,x1,y1);
}
