/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "ThemeConfig.h"

#include <stdlib.h>
#include <limits.h>

#include <QString>
#include <QVariant>
#include <QSettings>
#include <QFile>
#include <QStringList>

ThemeConfig::ThemeConfig() :
  m_valid(false),
  settings(NULL)
{
}

ThemeConfig::ThemeConfig(const QString& theme) :
  m_valid(false),
  settings(NULL)
{
  load(theme);
}

ThemeConfig::~ThemeConfig()
{
  if (settings) {
    settings->sync();
    delete settings;
  }
}

void ThemeConfig::load(const QString& theme)
{
  if (settings) {
    settings->sync();
    delete settings;
  }

  settings = NULL;
  m_valid = false;

  if (!QFile::exists(theme))
    return;

  settings = new QSettings(theme,QSettings::NativeFormat);

  m_valid = true;
}

QVariant ThemeConfig::getValue(const QString& group, const QString& key, bool* ok) const
{
  QVariant r;
  if (settings) {
    settings->beginGroup(group);
    if (ok)
      *ok = settings->contains(key);
    r = settings->value(key);
    settings->endGroup();
  }

  return r;
}

void ThemeConfig::setValue(const QString& group, const QString& key, const QString& v) const
{
  if (settings) {
    settings->beginGroup(group);
    settings->setValue(key,v);
    settings->endGroup();
  }
}

frame_spec_t ThemeConfig::getFrameSpec(const QString &widgetName) const
{
  frame_spec_t r;
  r.hasFrame = false;
  r.hasCapsule = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.rtop = r.rbottom = r.rleft = r.rright = false;
  r.ptop = r.pbottom = r.pleft = r.pright = 0;
  r.capsuleH = r.capsuleV = 0;

  bool ok;

  QVariant v = getValue(widgetName,"frame", &ok);
  if (ok) {
    r.hasFrame = v.toBool();
    if (r.hasFrame) {
      v = getValue(widgetName, "frame.element", &ok);
      if (ok)
        r.element = v.toString();
      v = getValue(widgetName,"frame.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"frame.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"frame.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"frame.right", &ok);
      if (ok)
        r.right = v.toInt();

      v = getValue(widgetName,"frame.capsule", &ok);
      if (ok)
        r.hasCapsule = v.toBool();

      v = getValue(widgetName,"frame.repeat.top",&ok);
      if (ok)
        r.rtop = v.toBool();
      v = getValue(widgetName,"frame.repeat.bottom",&ok);
      if (ok)
        r.rbottom = v.toBool();
      v = getValue(widgetName,"frame.repeat.left",&ok);
      if (ok)
        r.rleft = v.toBool();
      v = getValue(widgetName,"frame.repeat.right",&ok);
      if (ok)
        r.rright = v.toBool();

      if (r.rtop) {
        v = getValue(widgetName,"frame.repeat.top.patternsize",&ok);
        if (ok)
          r.ptop = v.toInt();
      }
      if (r.rbottom) {
        v = getValue(widgetName,"frame.repeat.bottom.patternsize",&ok);
        if (ok)
          r.pbottom = v.toInt();
      }
      if (r.rleft) {
        v = getValue(widgetName,"frame.repeat.left.patternsize",&ok);
        if (ok)
          r.pleft = v.toInt();
      }
      if (r.rright) {
        v = getValue(widgetName,"frame.repeat.right.patternsize",&ok);
        if (ok)
          r.pright = v.toInt();
      }
    }
  }

  return r;
}

void ThemeConfig::setFrameSpec(const QString& widgetName, const frame_spec_t &fspec) const
{
  setValue(widgetName, "frame", fspec.hasFrame ? "true" : "false");
  setValue(widgetName, "frame.element", fspec.element);
  setValue(widgetName, "frame.top", QString("%1").arg(fspec.top));
  setValue(widgetName, "frame.bottom", QString("%1").arg(fspec.bottom));
  setValue(widgetName, "frame.left", QString("%1").arg(fspec.left));
  setValue(widgetName, "frame.right", QString("%1").arg(fspec.right));
  setValue(widgetName, "frame.capsule", fspec.hasCapsule ? "true" : "false");
  setValue(widgetName, "frame.repeat.top", fspec.rtop ? "true" : "false");
  setValue(widgetName, "frame.repeat.bottom", fspec.rbottom ? "true" : "false");
  setValue(widgetName, "frame.repeat.left", fspec.rleft ? "true" : "false");
  setValue(widgetName, "frame.repeat.right", fspec.rright ? "true" : "false");
  setValue(widgetName, "frame.repeat.top.patternsize", QString("%1").arg(fspec.ptop));
  setValue(widgetName, "frame.repeat.bottom.patternsize", QString("%1").arg(fspec.pbottom));
  setValue(widgetName, "frame.repeat.left.patternsize", QString("%1").arg(fspec.pleft));
  setValue(widgetName, "frame.repeat.right.patternsize", QString("%1").arg(fspec.pright));
}

interior_spec_t ThemeConfig::getInteriorSpec(const QString &widgetName) const
{
  interior_spec_t r;
  r.hasInterior = true;
  r.hasMargin = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.rx = r.ry = false;
  r.px = r.py = 0;

  bool ok;

  QVariant v = getValue(widgetName,"interior", &ok);
  if (ok)
    r.hasInterior = v.toBool();

  if (r.hasInterior) {
    v = getValue(widgetName, "interior.element", &ok);
      if (ok)
        r.element = v.toString();
  }

  v = getValue(widgetName,"interior.margin", &ok);
  if (ok) {
    r.hasMargin = v.toBool();
    if (r.hasMargin) {
      v = getValue(widgetName,"interior.margin.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"interior.margin.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"interior.margin.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"interior.margin.right", &ok);
      if (ok)
        r.right = v.toInt();
    }
  }

  v = getValue(widgetName,"interior.repeat.x",&ok);
  if (ok)
    r.rx = v.toBool();
  v = getValue(widgetName,"interior.repeat.y",&ok);
  if (ok)
    r.ry = v.toBool();

  if (r.rx) {
    v = getValue(widgetName,"interior.repeat.x.patternsize",&ok);
    if (ok)
      r.px = v.toInt();
  }

  if (r.ry) {
    v = getValue(widgetName,"interior.repeat.y.patternsize",&ok);
    if (ok)
      r.py = v.toInt();
  }

  return r;
}

void ThemeConfig::setInteriorSpec(const QString& widgetName, const interior_spec_t& ispec) const
{
  setValue(widgetName, "interior", ispec.hasInterior ? "true" : "false");
  setValue(widgetName, "interior.element", ispec.element);
  setValue(widgetName, "interior.margin", ispec.hasMargin ? "true" : "false");
  setValue(widgetName, "interior.margin.top", QString("%1").arg(ispec.top));
  setValue(widgetName, "interior.margin.bottom", QString("%1").arg(ispec.bottom));
  setValue(widgetName, "interior.margin.left", QString("%1").arg(ispec.left));
  setValue(widgetName, "interior.margin.right", QString("%1").arg(ispec.right));
  setValue(widgetName, "interior.repeat.x", ispec.rx ? "true" : "false");
  setValue(widgetName, "interior.repeat.y", ispec.ry ? "true" : "false");
  setValue(widgetName, "interior.repeat.x.patternsize", QString("%1").arg(ispec.px));
  setValue(widgetName, "interior.repeat.y.patternsize", QString("%1").arg(ispec.px));
}

indicator_spec_t ThemeConfig::getIndicatorSpec(const QString &widgetName) const
{
  indicator_spec_t r;
  r.size = 15;

  bool ok;

  QVariant v = getValue(widgetName,"indicator.size", &ok);
  if (ok)
    r.size = v.toInt();


  v = getValue(widgetName, "indicator.element", &ok);
    if (ok)
      r.element = v.toString();

  return r;
}

void ThemeConfig::setIndicatorSpec(const QString& widgetName, const indicator_spec_t& dspec) const
{
  setValue(widgetName, "indicator.element", dspec.element);
  setValue(widgetName, "indicator.size", QString("%1").arg(dspec.size));
}

label_spec_t ThemeConfig::getTextSpec(const QString &widgetName) const
{
  label_spec_t r;
  r.hasShadow = false;
  r.xshift = 0;
  r.yshift = 1;
  r.r = r.g = r.b = 0;
  r.a = 255;
  r.depth = 1;
  r.hasMargin = false;
  r.top = r.bottom = r.left = r.right = 0;
  r.tispace = 0;

  bool ok;

  QVariant v = getValue(widgetName,"text.shadow", &ok);
  if (ok) {
    r.hasShadow = v.toBool();

    if (r.hasShadow) {
      v = getValue(widgetName,"text.shadow.xshift", &ok);
      if (ok)
        r.xshift = v.toInt();
      v = getValue(widgetName,"text.shadow.yshift", &ok);
      if (ok)
        r.yshift = v.toInt();
      v = getValue(widgetName,"text.shadow.red", &ok);
      if (ok)
        r.r = v.toInt();
      v = getValue(widgetName,"text.shadow.green", &ok);
      if (ok)
        r.g = v.toInt();
      v = getValue(widgetName,"text.shadow.blue", &ok);
      if (ok)
        r.b = v.toInt();
      v = getValue(widgetName,"text.shadow.alpha", &ok);
      if (ok)
        r.a = v.toInt();
      v = getValue(widgetName,"text.shadow.depth", &ok);
      if (ok)
        r.depth = v.toInt();
    }
  }

  v = getValue(widgetName,"text.margin", &ok);
  if (ok) {
    r.hasMargin = v.toBool();
    if (r.hasMargin) {
      v = getValue(widgetName,"text.margin.top", &ok);
      if (ok)
        r.top = v.toInt();
      v = getValue(widgetName,"text.margin.bottom", &ok);
      if (ok)
        r.bottom = v.toInt();
      v = getValue(widgetName,"text.margin.left", &ok);
      if (ok)
        r.left = v.toInt();
      v = getValue(widgetName,"text.margin.right", &ok);
      if (ok)
        r.right = v.toInt();
    }
  }

  v = getValue(widgetName,"text.iconspacing", &ok);
  if (ok)
    r.tispace = v.toInt();

  return r;
}

void ThemeConfig::setTextSpec(const QString& widgetName, const label_spec_t & lspec) const
{
  setValue(widgetName, "text.shadow", lspec.hasShadow ? "true" : "false");
  setValue(widgetName, "text.shadow.xshift", QString("%1").arg(lspec.xshift));
  setValue(widgetName, "text.shadow.yshift", QString("%1").arg(lspec.yshift));
  setValue(widgetName, "text.shadow.red", QString("%1").arg(lspec.r));
  setValue(widgetName, "text.shadow.green", QString("%1").arg(lspec.g));
  setValue(widgetName, "text.shadow.blue", QString("%1").arg(lspec.b));
  setValue(widgetName, "text.shadow.alpha", QString("%1").arg(lspec.a));
  setValue(widgetName, "text.margin", lspec.hasMargin ? "true" : "false");
  setValue(widgetName, "text.margin.top", QString("%1").arg(lspec.top));
  setValue(widgetName, "text.margin.bottom", QString("%1").arg(lspec.bottom));
  setValue(widgetName, "text.margin.left", QString("%1").arg(lspec.left));
  setValue(widgetName, "text.margin.right", QString("%1").arg(lspec.right));
  setValue(widgetName, "text.iconspacing", QString("%1").arg(lspec.tispace));
}

size_spec_t ThemeConfig::getSizeSpec(const QString& widgetName) const
{
  size_spec_t r;
  r.minH = r.fixedH = r.minW = r.fixedW = -1;

  bool ok;

  QVariant v = getValue(widgetName,"size.minheight", &ok);
  if (ok)
    r.minH = v.toInt();
  v = getValue(widgetName,"size.fixedheight", &ok);
  if (ok)
    r.fixedH = v.toInt();

  v = getValue(widgetName,"size.minwidth", &ok);
  if (ok)
    r.minW = v.toInt();
  v = getValue(widgetName,"size.fixedwidth", &ok);
  if (ok)
    r.fixedW = v.toInt();

  return r;
}

void ThemeConfig::setSizeSpec(const QString& widgetName, const size_spec_t &sspec) const
{
  setValue(widgetName, "size.minheight", QString("%1").arg(sspec.minH));
  setValue(widgetName, "size.fixedheight", QString("%1").arg(sspec.fixedH));
  setValue(widgetName, "size.minwidth", QString("%1").arg(sspec.minW));
  setValue(widgetName, "size.fixedwidth", QString("%1").arg(sspec.fixedW));
}


widget_spec_t ThemeConfig::getWidgetSpec(const QString& widgetName) const
{
  widget_spec_t r;
  r.frame = getFrameSpec(widgetName);
  r.interior = getInteriorSpec(widgetName);
  r.indicator = getIndicatorSpec(widgetName);
  r.label = getTextSpec(widgetName);
  r.size = getSizeSpec(widgetName);

  return r;
}

void ThemeConfig::setWidgetSpec(const QString& widgetName, const widget_spec_t& wspec) const
{
  setFrameSpec(widgetName,wspec.frame);
  setInteriorSpec(widgetName,wspec.interior);
  setIndicatorSpec(widgetName,wspec.indicator);
  setTextSpec(widgetName,wspec.label);
  setSizeSpec(widgetName,wspec.size);
}

QStringList ThemeConfig::getWidgets() const
{
  if (m_valid)
    return settings->childGroups();
  else
    return QStringList();
}
